/*
 * Copyright (c) 2012 Mans Rullgard <mans@mansr.com>
 *
 * This file is part of FFmpeg.
 *
 * FFmpeg is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * FFmpeg is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with FFmpeg; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */

#include "config.h"
#include "flacencdsp.h"

#define SAMPLE_SIZE 16
#include "flacdsp_lpc_template.c"

#undef  SAMPLE_SIZE
#define SAMPLE_SIZE 32
#include "flacdsp_lpc_template.c"

#define ZIGZAG_32BIT_MAX  0x3FFFFFFF
#define ZIGZAG_32BIT_MIN -0x3FFFFFFF

int ff_flacdsp_lpc_encode_c_32_overflow_detect(int32_t *res, const int32_t *smp, int len,
                                               int order, int32_t *coefs, int shift)
{
    /* This function checks for every prediction and every residual
     * whether they cause integer overflow in existing decoders. In
     * case the prediction exceeds int32_t limits, prediction
     * coefficients are lowered accordingly. If the residual exceeds
     * ZIGZAG_32BIT_MAX and _MIN or coefficients have been lowered
     * twice but the prediction still overflows, give up */
    int lpc_reduction_tries = 0;
    int64_t pmax;
    for (int i = 0; i < order; i++)
        res[i] = smp[i];
    do {
        pmax = 0;
        for (int i = order; i < len; i++) {
            int64_t p = 0, tmp;
            for (int j = 0; j < order; j++)
                p += (int64_t)coefs[j]*smp[(i-1)-j];
            p >>= shift;
            tmp = smp[i] - p;
            if (p > INT32_MAX && p > pmax)
                pmax = p;
            else if (p < INT32_MIN && (p * -1) > pmax)
                pmax = p * -1;
            if (tmp > ZIGZAG_32BIT_MAX || tmp < ZIGZAG_32BIT_MIN)
                return 0;
            res[i] = tmp;
        }

        if (pmax > 0) {
            if (lpc_reduction_tries >= 2)
                return 0;
            lpc_reduction_tries++;
            for (int i = 0; i < order; i++)
                coefs[i] = ((int64_t)coefs[i] * INT32_MAX) / pmax;
        }
    } while (pmax > 0);
    return 1;
}

av_cold void ff_flacencdsp_init(FLACEncDSPContext *c)
{
    c->lpc16_encode = flac_lpc_encode_c_16;
    c->lpc32_encode = flac_lpc_encode_c_32;

#if ARCH_X86
    ff_flacencdsp_init_x86(c);
#endif
}
